from quoridor.square import *
from quoridor.exceptions import *
from quoridor.pawn import Pawn
from quoridor.utils import BoolGrid
from collections import deque


class Board:
    def __init__(self):
        # Set up board
        self.board = [[Square(row, col) for col in range(MAX_COL)] for row in
                      range(MAX_ROW)]

        # Place neighbours
        for row in range(MAX_ROW):
            for col in range(MAX_COL):
                left_square = self.get_square_or_none(row, col - 1)
                right_square = self.get_square_or_none(row, col + 1)
                up_square = self.get_square_or_none(row + 1, col)
                down_square = self.get_square_or_none(row - 1, col)
                neighbours = {s for s in [up_square, right_square, down_square, left_square] if s is not None}
                self[(row, col)].set_physical_neighbours(neighbours)
                self[(row, col)].add_neighbours(neighbours)

        # Initialise fences. These are utilised to check whether fences exist
        self.horizontal_fence_grid = BoolGrid(MAX_ROW, MAX_COL)
        self.vertical_fence_grid = BoolGrid(MAX_ROW, MAX_COL)
        self.fence_center_grid = BoolGrid(MAX_ROW, MAX_COL)

        # Initialise pawns
        self.pawns = {'white': Pawn(square=self[(WHITE_INIT_ROW, WHITE_INIT_COL)], color='white'),
                      # WHITE_INIT_ROW, WHITE_INIT_COL
                      'black': Pawn(square=self[(BLACK_INIT_ROW, BLACK_INIT_COL)],
                                    color='black')}  # BLACK_INIT_ROW, BLACK_INIT_COL

        # Update possible moves based on initial pawn positions
        self.update_neighbours(self.black_pawn.square)
        self.update_neighbours(self.white_pawn.square)

        # Initialise grids indicating whether there is a path for every player
        self.path_exists = {}
        for player in self.pawns.keys():
            self.path_exists[player] = BoolGrid(MAX_ROW, MAX_COL, init_value=True)

    @property
    def white_pawn(self):
        return self.pawns['white']

    @property
    def black_pawn(self):
        return self.pawns['black']

    def check_winner(self):
        for player, pawn in self.pawns.items():
            if pawn.square.winning_square(player):
                raise GameOver(player, last_move=False, message=f'Player {player} wins the game')

    def game_finished(self):
        try:
            self.check_winner()
            return False
        except GameOver:
            return True

    def partial_FEN(self):
        """
        Format: [1] / [2] / [3.1] [3.2] [3.3*] [3.4*] / [4.1] [4.2] [4.3*] [4.4*] / [5]
        1. Horizontal wall positions
        2. Vertical wall positions
        3. Pawn positions:
            3.1 Player 1 pawn position
            3.2 Player 2 pawn position
            3.3 Player 3 pawn position*
            3.4 Player 4 pawn position*
        4. Walls available: (not included here)
            4.1 player 1 walls available
            4.2 player 2 walls available
            4.3 player 3 walls available*
            4.4 player 4 walls available*
        5. Active player (not included here)
        * Four player only.
        """
        # Walls
        vertical_fences = sorted(self.vertical_fence_grid.argwhere_str())
        horizontal_fences = sorted(self.horizontal_fence_grid.argwhere_str(), key=lambda x: x[1])
        # Remove even positions in lists -- only first coordinate is used in the FEN
        del vertical_fences[1::2]
        del horizontal_fences[1::2]

        # Pawn positions and walls available
        white_position = self.pawns['white'].square.square2str()
        black_position = self.pawns['black'].square.square2str()
        pawn_positions = white_position + black_position

        # Construct FEN
        fen = f'{"".join(sorted(horizontal_fences))}/{"".join(sorted(vertical_fences))}/{pawn_positions}'
        return fen

    def from_FEN(self, FEN):
        raise NotImplementedError('From FEN not implemented')

    def __getitem__(self, location):
        row, col = location
        # Check whether new square is valid
        if not Square.valid_square(row, col):
            raise InvalidSquare(f'Invalid square ({row}, {col})')

        return self.board[row][col]

    def get_square_or_none(self, row, col):
        try:
            square = self[(row, col)]
        except InvalidSquare:
            square = None
        return square

    def legal_moves(self):
        raise NotImplementedError()

    def fence_exists(self, row, col, orientation):
        if orientation == 'v':
            return self.vertical_fence_grid[(row, col)] \
                   or self.vertical_fence_grid[(row + 1, col)] \
                   or self.fence_center_grid[(row, col)]
        return self.horizontal_fence_grid[(row, col)] \
               or self.horizontal_fence_grid[(row, col + 1)] \
               or self.fence_center_grid[(row, col)]

    def _set_pawn_location(self, player, target_row, target_col):
        target_square = self[(target_row, target_col)]
        self.pawns[player].move(target_square)

        # Reset neighbours of all pawns
        for pawn in self.pawns.values():
            pawn.square.reset_neighbours()

        # Update neighbours of all pawns
        for pawn in self.pawns.values():
            self.update_neighbours(pawn.square)

    def _add_or_remove_neighbours(self, source_row, source_col, target_row, target_col, add):
        """
        Removes mutual neighbours (source, target) in neighbours and physical neighbours
        """
        source_square = self.get_square_or_none(source_row, source_col)
        target_square = self.get_square_or_none(target_row, target_col)
        if source_square is not None and target_square is not None:
            source_square.add_or_remove_neighbours(target_square, add=add)
            target_square.add_or_remove_neighbours(source_square, add=add)
            source_square.add_or_remove_physical_neighbours(target_square, add=add)
            target_square.add_or_remove_physical_neighbours(source_square, add=add)

    def _remove_diagonal_neighbours(self, square):
        """
        Removes mutual neighbours (source, target) in neighbours and physical neighbours
        """
        self._add_or_remove_neighbours(square.row, square.col, square.row + 1, square.col - 1, add=False)
        self._add_or_remove_neighbours(square.row, square.col, square.row + 1, square.col + 1, add=False)
        self._add_or_remove_neighbours(square.row, square.col, square.row - 1, square.col - 1, add=False)
        self._add_or_remove_neighbours(square.row, square.col, square.row - 1, square.col + 1, add=False)

    def _pawn_conditional_remove_neighbours(self, source_row, source_col,
                                            target_row, target_col,
                                            cond_row, cond_col):
        """
        Adds or removes a particular neighbourhood connection between source and target only if there is a pawn
        in the cond square.
        """
        target_square = self.get_square_or_none(target_row, target_col)
        cond_square = self.get_square_or_none(cond_row, cond_col)
        if target_square is not None and cond_square is not None and not cond_square.has_pawn():
            self._add_or_remove_neighbours(source_row, source_col, target_row, target_col, add=False)

    def _place_or_remove_fence(self, row, col, orientation, place=True):
        """
        Places or removes a fence and updates the square neighbourhoods accordingly.
        """
        # Assumes valid fence
        if orientation == 'h':
            self._add_or_remove_neighbours(row, col, row + 1, col, add=not place)
            self._add_or_remove_neighbours(row, col + 1, row + 1, col + 1, add=not place)

            # Remove distant straight neighbours (pawn in between)
            if place:
                self._add_or_remove_neighbours(row, col, row + 2, col, add=False)
                self._add_or_remove_neighbours(row, col + 1, row + 2, col + 1, add=False)
                self._add_or_remove_neighbours(row + 1, col, row - 1, col, add=False)
                self._add_or_remove_neighbours(row + 1, col + 1, row - 1, col + 1, add=False)

                # Remove distant diagonal neighbours (pawn in between)
                self._pawn_conditional_remove_neighbours(row, col, row + 1, col - 1, row, col - 1)
                self._pawn_conditional_remove_neighbours(row + 1, col, row, col - 1, row + 1, col - 1)
                self._pawn_conditional_remove_neighbours(row, col + 1, row + 1, col + 2, row, col + 2)
                self._pawn_conditional_remove_neighbours(row + 1, col + 1, row, col + 2, row + 1, col + 2)
                self._add_or_remove_neighbours(row, col + 1, row + 1, col, add=False)
                self._add_or_remove_neighbours(row + 1, col + 1, row, col, add=False)

            # Place fence
            self.horizontal_fence_grid[(row, col)] = place
            self.horizontal_fence_grid[(row, col + 1)] = place
            self.fence_center_grid[(row, col)] = place

        else:  # Vertical fence
            self._add_or_remove_neighbours(row, col, row, col + 1, add=not place)
            self._add_or_remove_neighbours(row + 1, col, row + 1, col + 1, add=not place)

            if place:
                # Remove distant straight neighbours (pawn in between)
                self._add_or_remove_neighbours(row, col, row, col + 2, add=False)
                self._add_or_remove_neighbours(row + 1, col, row + 1, col + 2, add=False)
                self._add_or_remove_neighbours(row, col + 1, row, col - 1, add=False)
                self._add_or_remove_neighbours(row + 1, col + 1, row + 1, col - 1, add=False)

                # Remove distant diagonal neighbours (pawn in between)
                self._pawn_conditional_remove_neighbours(row - 1, col, row, col + 1, row - 1, col + 1)
                self._pawn_conditional_remove_neighbours(row - 1, col + 1, row, col, row - 1, col)
                self._pawn_conditional_remove_neighbours(row + 1, col, row + 2, col + 1, row + 2, col)
                self._pawn_conditional_remove_neighbours(row + 1, col + 1, row + 2, col, row + 2, col + 1)
                self._add_or_remove_neighbours(row + 1, col, row, col + 1, add=not place)
                self._add_or_remove_neighbours(row, col, row + 1, col + 1, add=not place)

            # Place fence
            self.vertical_fence_grid[(row, col)] = place
            self.vertical_fence_grid[(row + 1, col)] = place
            self.fence_center_grid[(row, col)] = place

        # Update connections of all pawns (accounts for straight and diagonal jumps)
        for _, pawn in self.pawns.items():
            self.update_neighbours(pawn.square)

    def BFS_all(self, player):
        """
        Breadth-first search from end squares to all squares in the board.
        """
        path_exists = BoolGrid(MAX_ROW, MAX_COL, init_value=False)
        visited_squares = BoolGrid(MAX_ROW, MAX_COL, init_value=False)
        square_queue = deque()

        # Set end row according to player
        end_row = 0
        if player == 'white':
            end_row = MAX_ROW - 1

        # Initialise paths, visited squares and queue
        for i in range(MAX_COL):
            path_exists[(end_row, i)] = True
            visited_squares[(end_row, i)] = True
            square_queue.append(self[(end_row, i)])

        # Keep popping neighbours until the queue is empty. All remaining unvisited squares have no path to the goal
        while len(square_queue) != 0:
            square = square_queue.popleft()
            for neighbour in square.neighbours_iter():
                row = neighbour.row
                col = neighbour.col
                if not visited_squares[(row, col)]:
                    path_exists[(row, col)] = True
                    square_queue.append(neighbour)
                    visited_squares[(row, col)] = True

        return path_exists

    def BFS_player(self, player):
        """
        Breadth-first search from player pawn square.
        """
        visited_squares = BoolGrid(MAX_ROW, MAX_COL, init_value=False)
        square_queue = deque()

        # Set end row according to player
        end_row = 0
        if player == 'white':
            end_row = MAX_ROW - 1

        # Initialise paths, visited squares and queue
        pawn_square = self.pawns[player].square
        visited_squares[(pawn_square.row, pawn_square.col)] = True
        square_queue.append(pawn_square)

        # Keep popping neighbours until the queue is empty. All remaining unvisited squares have no path to the goal
        can_reach = False
        while len(square_queue) != 0 and not can_reach:
            square = square_queue.popleft()
            for neighbour in square.neighbours_iter():
                row = neighbour.row
                col = neighbour.col
                if row == end_row:
                    can_reach = True
                    break
                if not visited_squares[(row, col)]:
                    square_queue.append(neighbour)
                    visited_squares[(row, col)] = True

        return can_reach

    def place_fence(self, row, col, orientation, check_winner=True, run_BFS=True):
        """
        Places a fence or raises InvalidFence exception if the fence is invalid.
        """
        if check_winner:
            self.check_winner()

        # Check if neighbouring squares are valid
        try:
            squares = [self[row, col],
                       self[row + 1, col],
                       self[row, col + 1],
                       self[row + 1, col + 1]]
        except InvalidSquare:
            raise InvalidFence(f'Invalid fence ({row}, {col}): Fence out of bounds')

        # Check if fence position is occupied
        invalid = self.fence_exists(row, col, orientation)
        if invalid:
            raise InvalidFence(f'Invalid fence ({row}, {col}): Fence overlap')

        # Place fence temporarily
        self._place_or_remove_fence(row, col, orientation, place=True)

        # Check that path exists for all pawns
        valid_fence = True
        if run_BFS:  # In some cases where BFS has been ran in another process, we don't want to repeat
            for player in self.pawns.keys():
                valid_fence = self.BFS_player(player)
                if not valid_fence:
                    break

        # Undo the fence if one of the players can't reach the end
        if not valid_fence:
            self._place_or_remove_fence(row, col, orientation, place=False)
            raise InvalidFence(f'Invalid fence ({row}, {col}): Fence blocks {player} pawn path')

    def _add_diagonal_skip_neighbours(self, square, next_square, orientation='h'):
        """
        When straight skip connection is not possible, adds skip diagonal connections (neighbours of square)
        to next_square
        """
        if square is not None and next_square is not None and \
                square.has_pawn() and next_square.has_neighbour(square):
            row = square.row
            col = square.col
            diff_row = 0
            diff_col = 1
            if orientation == 'v':
                diff_row = 1
                diff_col = 0

            # Add diagonal skip neighbours to next_square
            neighbour_1 = self.get_square_or_none(row + diff_row, col + diff_col)
            neighbour_2 = self.get_square_or_none(row - diff_row, col - diff_col)
            if neighbour_1 is not None and square.has_neighbour(neighbour_1):
                next_square.add_neighbours(neighbour_1)
            if neighbour_2 is not None and square.has_neighbour(neighbour_2):
                next_square.add_neighbours(neighbour_2)

    def update_neighbours(self, target_square):
        """
        Updates neighbours of a target square occupied by a pawn
        :param target_square: target square for which to update neighbours
        :return squares of neighbour squares occupied by pawns
        """
        occupied_squares = []

        # Update neighbours of final square
        for square in target_square.physical_neighbours_iter():
            # Increasing neighbourhood radius
            if square.has_pawn():  # square.is_neighbour(target_square) and
                occupied_squares.append(square)

                # Make jump connections
                diff_row = square.row - target_square.row
                diff_col = square.col - target_square.col
                next_square = self.get_square_or_none(square.row + diff_row, square.col + diff_col)

                # No fence behind pawn
                if next_square is not None and square.has_neighbour(next_square):
                    target_square.add_neighbours(next_square)
                    # square.add_neighbours(source_square)
                else:  # Fence behind the pawn. Get two other neighbours
                    self._add_diagonal_skip_neighbours(square=square,
                                                       next_square=target_square,
                                                       orientation='h')
                    self._add_diagonal_skip_neighbours(square=square,
                                                       next_square=target_square,
                                                       orientation='v')
        return occupied_squares

    def move_pawn(self, player, target_row, target_col, check_winner=True):
        """
        Moves a pawn or raises InvalidMove exception if the move is invalid.
        """
        if check_winner:
            self.check_winner()

        pawn = self.pawns[player]
        source_square = pawn.square
        target_square = self[(target_row, target_col)]
        if target_square.has_pawn() or not source_square.has_neighbour(target_square):
            raise InvalidMove(f'Cannot move {player} pawn to square ({target_row}, {target_col})')

        # Execute transaction: Move pawn
        pawn.move(target_square)

        # Check if game is over
        if target_square.winning_square(player=player):
            raise GameOver(player, last_move=True, message=f'Player {player} wins the game')

        # Remove diagonal neighbours of final square
        # self._remove_diagonal_neighbours(target_square)

        # Reset neighbours of all pawns
        for pawn in self.pawns.values():
            pawn.square.reset_neighbours()

        # Update neighbours of all pawns
        for pawn in self.pawns.values():
            self.update_neighbours(pawn.square)

    def valid_pawn_moves(self, player, check_winner=True):
        """
        Returns a list of valid move squares of a pawn.
        """
        if check_winner:
            self.check_winner()

        pawn = self.pawns[player]
        # return pawn.square.neighbours
        return {s for s in pawn.square.neighbours if not s.has_pawn()}



from quoridor.exceptions import InvalidSquare
from quoridor.utils import coord2str

# Square constants
MIN_ROW = 0
MIN_COL = 0
MAX_ROW = 9
MAX_COL = 9

WHITE_INIT_ROW = 0
WHITE_INIT_COL = 4
BLACK_INIT_ROW = 8
BLACK_INIT_COL = 4

DIRECTIONS = {'down': 0, 'right': 1, 'up': 2, 'left': 3}
OPPOSITE_DIRECTIONS = {'down': 'up', 'right': 'left', 'up': 'down', 'left': 'right'}


class Square:
    def __init__(self, row, col, pawn=None):
        if not self.valid_square(row, col):
            raise InvalidSquare(f'Invalid square: ({row}, {col})')
        self.row = row
        self.col = col
        self.pawn = pawn
        self.neighbours = set()
        self.physical_neighbours = set()  # [None, None, None, None]  # Up, right, down, left

    def set_physical_neighbours(self, neighbours):
        self.physical_neighbours = neighbours

    def physical_neighbours_iter(self):
        for square in self.physical_neighbours:
            if square is not None:
                yield square

    def neighbours_iter(self):
        for square in self.neighbours:
            if square is not None:
                yield square

    def add_or_remove_neighbours(self, neighbours, add=True):
        try:
            iter(neighbours)
        except TypeError:
            neighbours = [neighbours]

        neighbour_set = set(neighbours)
        if add:
            self.neighbours = self.neighbours.union(neighbour_set)
        else:
            self.neighbours = self.neighbours.difference(neighbour_set)

    def add_neighbours(self, neighbours):
        self.add_or_remove_neighbours(neighbours, add=True)

    def remove_neighbours(self, neighbours):
        self.add_or_remove_neighbours(neighbours, add=False)

    def reset_neighbours(self):
        self.neighbours = self.physical_neighbours.copy()

    def add_or_remove_physical_neighbours(self, neighbours, add=True):
        try:
            iter(neighbours)
        except TypeError:
            neighbours = [neighbours]

        neighbour_set = set(neighbours)
        if add:
            self.physical_neighbours = self.physical_neighbours.union(neighbour_set)
        else:
            self.physical_neighbours = self.physical_neighbours.difference(neighbour_set)

    def add_physical_neighbours(self, neighbours):
        self.add_or_remove_physical_neighbours(neighbours, add=True)

    def remove_physical_neighbours(self, neighbours):
        self.add_or_remove_physical_neighbours(neighbours, add=False)

    def has_pawn(self):
        return self.pawn is not None

    @property
    def location(self):
        return self.row, self.col

    @staticmethod
    def valid_square(x, y):
        return (MIN_ROW <= x < MAX_ROW) & (MIN_COL <= y < MAX_COL)

    def adjacent_coords(self, direction):
        x, y = self.row, self.col
        new_x, new_y = x, y
        if direction == 'right':
            new_x = x + 1
        if direction == 'up':
            new_y = y + 1
        if direction == 'left':
            new_x = x - 1
        if direction == 'down':
            new_y = y - 1
        return new_x, new_y

    def winning_square(self, player='white'):
        if player == 'white':
            return self.row == MAX_ROW - 1
        else:
            return self.row == MIN_ROW

    def set_pawn(self, pawn):
        self.pawn = pawn

    def is_busy(self):
        return self.pawn is not None

    def square2str(self):
        return coord2str(self.row, self.col)

    def __str__(self):
        return f'{self.row}, {self.col}'

    def __hash__(self):
        return hash((self.row, self.col))

    def __eq__(self, other):
        return (self.row, self.col) == (other.row, other.col)

    def __ne__(self, other):
        # Not strictly necessary, but to avoid having both x==y and x!=y
        # True at the same time
        return not (self == other)

    def has_neighbour(self, other):
        return other in self.neighbours



class InvalidSquare(Exception):
    def __init__(self, message):
        super().__init__(message)


class InvalidMove(Exception):
    def __init__(self, message):
        super().__init__(message)


class InvalidFence(Exception):
    def __init__(self, message):
        super().__init__(message)


class TrappedPawn(Exception):
    def __init__(self, message):
        super().__init__(message)


class GameOver(Exception):
    def __init__(self, winner, last_move, message):
        super().__init__(message)
        self.last_move = last_move
        self.winner = winner

class InvalidPlayer(Exception):
    def __init__(self, message):
        super().__init__(message)


import numpy as np


class BoolGrid:
    def __init__(self, max_x, max_y, init_value=False):
        self.max_x = max_x
        self.max_y = max_y
        self.grid = []
        for x in range(max_x):
            row = [init_value] * max_y
            self.grid.append(row)
        self.grid = np.array(self.grid)

    def __getitem__(self, location):
        row, col = location
        return self.grid[row][col]

    def __setitem__(self, location, value):
        row, col = location
        self.grid[row][col] = value

    def argwhere(self):
        return np.argwhere(self.grid)

    def argwhere_str(self):
        return [coord2str(row, col) for row, col in self.argwhere()]


def coord2str(row, col):
    col_char = chr(col + 65)
    row = row + 1
    return f'{col_char}{row}'


def print_board(board, player_path_check=None):
    if player_path_check is not None:
        path_exists = board.BFS_all(player_path_check)

    nx = len(board.board)
    ny = len(board.board)
    for y in range(ny - 1, -1, -1):
        # Fence: print(' ---' * nx)
        row_fence_str = '  '

        row_str = str(y) + ' :'
        for x in range(nx):
            # Horizontal fences
            # down_fence = square.get_fence('up')
            if board.fence_center_grid[(y, x - 1)]:
                row_fence_str += '*'
            else:
                row_fence_str += ' '

            if board.horizontal_fence_grid[(y, x)]:  # down_fence is None:
                row_fence_str += '---'
            else:
                row_fence_str += '···'

            # Vertical fences
            # right_fence = square.get_fence('right')
            if board.vertical_fence_grid[(y, x)]:
                right_str = '|'
            else:
                right_str = ':'

            # print(':   ' * nx + ':')
            path_check = ' '
            if player_path_check is not None:
                if path_exists[(y, x)]:
                    # content = ' + '
                    path_check = '+'
                else:
                    # content = ' - '
                    path_check = '-'
            content = f' {path_check} '
            if board[y, x].is_busy():
                pawn_color = board[y, x].pawn.color
                content = f'{path_check}{pawn_color[0]} '
            row_str += f'{content}{right_str}'
        print(row_fence_str)
        # print('  ' + ' ···' * nx)
        print(row_str)
        # Fence: print('|   ' * nx + '|')
    print('  ' + ' ···' * nx)
    print('  ' + ''.join([f'  {i} ' for i in range(nx)]))

# def print_board_path_check(board, player):



from quoridor.board import Board


def board_setup1():
    board = Board()
    board.place_fence(row=0, col=0, orientation='h')
    board.place_fence(row=7, col=0, orientation='h')
    board.place_fence(row=3, col=2, orientation='v')
    board.place_fence(row=1, col=0, orientation='v')
    board._set_pawn_location('white', target_row=3, target_col=4)
    board._set_pawn_location('black', target_row=5, target_col=4)
    return board



from quoridor.board import Board
from quoridor.utils import print_board


# Press the green button in the gutter to run the script.
if __name__ == '__main__':
    board = Board()

    board.place_fence(row=0, col=0, orientation='h')
    board.place_fence(row=7, col=0, orientation='h')
    board.place_fence(row=3, col=2, orientation='v')
    board.place_fence(row=1, col=0, orientation='v')

    board.move_pawn(player='white', target_row=1, target_col=4)
    print_board(board)





import unittest
from quoridor.utils import print_board
from quoridor.exceptions import *
from quoridor.board_setups import *


class BoardIntegrity(unittest.TestCase):
    def test_invalid_fences(self):
        # Board set up
        board = board_setup1()

        # Overlapping horizontal fences
        with self.assertRaises(InvalidFence):
            board.place_fence(row=1, col=0, orientation='h')

        # Overlapping vertical fences
        with self.assertRaises(InvalidFence):
            board.place_fence(row=3, col=2, orientation='h')

        # Horizontal fence out of board
        with self.assertRaises(InvalidFence):
            board.place_fence(row=8, col=1, orientation='h')

        # Vertical fence out of board
        with self.assertRaises(InvalidFence):
            board.place_fence(row=2, col=8, orientation='v')

        # Horizontal fence on edge of the board
        with self.assertRaises(InvalidFence):
            board.place_fence(row=4, col=8, orientation='h')

        # Vertical fence on edge of the board
        with self.assertRaises(InvalidFence):
            board.place_fence(row=8, col=4, orientation='v')

        # Vertical edge crossing horizontal edge
        with self.assertRaises(InvalidFence):
            board.place_fence(row=0, col=0, orientation='v')

        # Horizontal edge crossing vertical edge
        with self.assertRaises(InvalidFence):
            board.place_fence(row=3, col=2, orientation='h')

    def test_invalid_moves(self):
        # Board set up
        board = board_setup1()

        # Fence down
        board.place_fence(row=2, col=4, orientation='h')
        with self.assertRaises(InvalidMove):
            board.move_pawn(player='white', target_row=2, target_col=4)

        board.move_pawn(player='white', target_row=4, target_col=4)
        board.move_pawn(player='white', target_row=6, target_col=4)

        # Fence left
        board.place_fence(row=5, col=3, orientation='v')
        with self.assertRaises(InvalidMove):
            board.move_pawn(player='white', target_row=6, target_col=3)

        # Fence up
        board.place_fence(row=6, col=4, orientation='h')
        with self.assertRaises(InvalidMove):
            board.move_pawn(player='white', target_row=7, target_col=4)

        # Fence right
        board.place_fence(row=5, col=4, orientation='v')
        with self.assertRaises(InvalidMove):
            board.move_pawn(player='white', target_row=6, target_col=5)

    def test_invalid_jump_moves(self):
        # Board set up
        board = board_setup1()
        board.move_pawn(player='black', target_row=4, target_col=4)
        board.move_pawn(player='black', target_row=2, target_col=4)
        board.move_pawn(player='black', target_row=2, target_col=3)
        board.move_pawn(player='white', target_row=3, target_col=3)
        board.move_pawn(player='white', target_row=4, target_col=3)
        # board.update_neighbours(board.black_pawn.square)
        # board.update_neighbours(board.white_pawn.square)

        # Move pawns
        board.move_pawn(player='white', target_row=3, target_col=3)
        board.move_pawn(player='white', target_row=1, target_col=3)
        with self.assertRaises(GameOver):
            board.move_pawn(player='black', target_row=0, target_col=3)

        # White pawn cannot jump over black pawn - fence in the middle
        board.place_fence(row=1, col=2, orientation='h', check_winner=False)
        with self.assertRaises(InvalidMove):
            board.move_pawn(player='black', target_row=2, target_col=3, check_winner=False)

        # Black pawn moves diagonally
        board.move_pawn(player='black', target_row=1, target_col=2, check_winner=False)

        # White pawn jumps black pawn
        board.move_pawn(player='white', target_row=1, target_col=1)

        # Black pawn straight jumps white pawn
        with self.assertRaises(InvalidMove):
            board.move_pawn(player='black', target_row=1, target_col=0)

        # Black pawn moves diagonally
        board.move_pawn(player='black', target_row=2, target_col=1)

        # White pawn tries to jump diagonally
        with self.assertRaises(InvalidMove):
            board.move_pawn(player='white', target_row=2, target_col=2)

        # White pawn jumps straight
        board.move_pawn(player='white', target_row=3, target_col=1)

        # Black pawn jumps diagonally
        board.place_fence(row=3, col=1, orientation='h')
        board.move_pawn(player='black', target_row=3, target_col=2)

        # White pawn jumps diagonally
        board.move_pawn(player='white', target_row=2, target_col=2)

        # Black pawn jumps diagonally
        board.move_pawn(player='black', target_row=2, target_col=3)
        # print_board(board)

    def test_more_invalid_moves(self):
        # Board set up
        board = Board()
        board._set_pawn_location('white', target_row=3, target_col=4)
        board._set_pawn_location('black', target_row=5, target_col=4)
        board.move_pawn(player='white', target_row=4, target_col=4)

        # White pawn can jump black. Move black pawn so that it's no longer possible
        board.move_pawn(player='black', target_row=5, target_col=5)
        with self.assertRaises(InvalidMove):
            board.move_pawn(player='white', target_row=6, target_col=4)

    def test_path_exists(self):
        # Board set up
        board = board_setup1()
        board.update_neighbours(board.black_pawn.square)
        board.update_neighbours(board.white_pawn.square)

        # Place fences to block path for black pawn
        board.place_fence(row=2, col=1, orientation='h')
        board.place_fence(row=2, col=3, orientation='h')
        board.place_fence(row=2, col=5, orientation='h')

        # Fence blocks black pawn's path
        with self.assertRaises(InvalidFence):
            board.place_fence(row=2, col=7, orientation='h')

        # Move pawn white upwards and place a fence that blocks white's path
        board.move_pawn(player='white', target_row=4, target_col=4)
        board.place_fence(row=5, col=2, orientation='v')
        board.place_fence(row=5, col=5, orientation='h')
        board.place_fence(row=5, col=7, orientation='h')

        # Fence blocks white pawn's path
        with self.assertRaises(InvalidFence):
            board.place_fence(row=5, col=3, orientation='h')

        # Check that the fence is undone properly, i.e. white pawn can jump over black
        board.move_pawn(player='white', target_row=6, target_col=4)
        board.move_pawn(player='white', target_row=4, target_col=4)
        board.move_pawn(player='white', target_row=4, target_col=5)
        board.move_pawn(player='white', target_row=5, target_col=5)

        # Check that we can put vertical fence that doesn't block the path
        board.place_fence(row=4, col=3, orientation='v')

        # Check that we can jump diagonally with white pawn
        board.move_pawn(player='white', target_row=6, target_col=4)

        # Check that we can now put the fence that separates white and black pawns
        board.place_fence(row=5, col=3, orientation='h')
        # print_board(board, player_path_check='black')

    def test_path_exists_2(self):
        # Board set up
        board = Board()

        # Place fences to block path for black pawn
        board.place_fence(row=5, col=0, orientation='h')
        board.place_fence(row=5, col=2, orientation='h')
        board.place_fence(row=5, col=4, orientation='h')
        board.place_fence(row=5, col=6, orientation='h')
        board.place_fence(row=5, col=7, orientation='v')
        # print_board(board, player_path_check='white')

        # Fence blocks black pawn's path
        with self.assertRaises(InvalidFence):
            board.place_fence(row=6, col=7, orientation='h')

        # Fence blocks black pawn's path
        with self.assertRaises(InvalidFence):
            # print([(s.row, s.col) for s in board[(5, 7)].neighbours])
            board.place_fence(row=4, col=7, orientation='h')
            # print_board(board, player_path_check='white')
        # print([(s.row, s.col) for s in board[(5, 7)].neighbours])
        # print([(s.row, s.col) for s in board[(4, 4)].neighbours])

        with self.assertRaises(InvalidFence):
            board.place_fence(row=4, col=7, orientation='h')

        print_board(board, player_path_check='white')
        # print(board.partial_FEN())


if __name__ == '__main__':
    unittest.main()



class Pawn:
    def __init__(self, square=None, color='white'):
        self.square = square
        self.color = color
        self.square.set_pawn(self)

    def move(self, square):
        self.square.set_pawn(None)
        self.square = square
        self.square.set_pawn(self)

    def wins(self):
        return self.square.winning_square(self.color)